#define VERSION "0.0.2"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#define BAUDRATE B115200
#define MODEMDEVICE "/dev/mp2000"
#define _POSIX_SOURCE 1		/* POSIX compliant source */
#define FALSE 0
#define TRUE 1

#define MODE0 (sendappcmd(fd,"@D----"),sendappcmd(fd, "@M0"))
#define MODE1 (sendappcmd(fd,"@D--M1"),sendappcmd(fd, "@M1"))
#define MODE2 (sendappcmd(fd,"@D--M2"),sendappcmd(fd, "@M2"))
#define MODE3 (sendappcmd(fd,"@D--M3"),sendappcmd(fd, "@M3"))
#define MODE4 (sendappcmd(fd,"@D--M4"),sendappcmd(fd, "@M4"))

volatile int STOP = FALSE;

unsigned char a2h(unsigned char hex)
{
	if (hex >= 0x30 && hex <= 0x39)
		return hex - 0x30;
	return hex - 0x41 + 10;
}

int sendbioscmd(int fd, char *cmd, char *answer)
{
	int res;
	int offset = 0;
	char buf[255];
#ifdef DEBUG_BIOS
	printf("sending cmd %s\n", cmd);
#endif
	write(fd, cmd, strlen(cmd));
	STOP = FALSE;
	while (STOP == FALSE) {
		res = read(fd, buf, 255);
		buf[res] = 0;
#ifdef DEBUG_BIOS
		printf("buf: -%s-\n", buf);
		printf("answer+off: -%s-\n", answer + offset);
		printf("strlen: %d strncmp: %d\n", strlen(buf), strncmp(buf, answer + offset, strlen(buf) - 1));
#endif
		if (strncmp(buf, answer + offset, strlen(buf) - 1) == 0) {
			offset += res;
			if (offset >= strlen(answer) - 1)
				STOP = TRUE;
		}
		else {
			fprintf(stderr, "Sending BIOS command %s did not produce expected answer %s\n", cmd, answer);
			return 0;
		}
	}
#ifdef DEBUG_BIOS
	printf("mp2k answer %s OK\n", answer);
#endif
	return 1;
}

unsigned char *getbiosinfo(int fd, char *cmd)
{
	static unsigned char answer[255];
	int res;
	char buf[255];
#ifdef DEBUG_BIOSINFO
	printf("sending cmd %s\n", cmd);
#endif
	write(fd, cmd, strlen(cmd));
	STOP = FALSE;
	answer[0] = 0;
	while (STOP == FALSE) {
		res = read(fd, buf, 255);
		buf[res] = 0;
#ifdef DEBUG_BIOSINFO
		printf("buf: -%s-\n", buf);
#endif
		strcat(answer, buf);
		if (buf[res - 1] == '>')
			STOP = TRUE;
	}
#ifdef DEBUG_BIOSINFO
	printf("mp2k answer is %s\n", answer);
#endif
	answer[strlen(answer) - 3] = 0;
	return answer;
}

unsigned char *sendappcmd(int fd, char *cmd)
{
	static unsigned char answer[255];
	int res = 0;
	int i;
	unsigned char length;
	unsigned char buf[255];
#ifdef DEBUG_APP
	printf("sending cmd %s\n", cmd);
#endif
	write(fd, cmd, strlen(cmd));
	while (!res)
		res = read(fd, &length, 1);
#ifdef DEBUG_APP
	printf("answer length: %d\n", length);
#endif
	answer[length] = 0;
	for (i = 0; i < length; i += res) {
		res = read(fd, buf, 255);
		memcpy(answer + i, buf, res);
	}
#ifdef DEBUG_APP
	printf("answer: %s\n", answer);
#endif
	return answer;
}

int checkwafer(fd)
{
	if (strcmp(sendappcmd(fd, "@K"), "03") == 0)
		return 1;
	else
		return 0;
}

int erasecard(int fd)
{
	if (!checkwafer(fd))
		return 0;
	sendappcmd(fd, "@DERAS");
	MODE1;
	sendappcmd(fd, "@ErA");
	MODE0;
	return 1;
}

#include "eeloader.h"

int loadeeprom(int fd, unsigned char *data)
{
	int counter = 0;
	int i;
	char cmdstring[255];

	MODE2;

	printf("Uploading EEPROM loader ");
	sendappcmd(fd, "@DLOAD");

	while (strcmp(eeloader[counter], "end")) {
		if (!(counter % 16)) {
			sprintf(cmdstring, "@DL%03X", counter);
			sendappcmd(fd, cmdstring);
			printf(".");
			fflush(stdout);
		}
		sprintf(cmdstring, "@L00%s", eeloader[counter]);
		if (strcmp(sendappcmd(fd, cmdstring), eeloader[counter])) {
			fprintf(stderr, "Upload error\n");
			return 0;
		}
		sendappcmd(fd, "@I");
		counter++;
	}

	printf("\nProgramming EEPROM ");
	MODE2;
	sendappcmd(fd, "@P003FFF");
	for (i = 0; i < 7; i++)
		sendappcmd(fd, "@I");
	sendappcmd(fd, "@L000001");
	if (strcmp(MODE4, "load")) {
		fprintf(stderr, "\nUnable to switch to EEPROM write mode\n");
		return 0;
	}

	for (i = 0; i < 2048; i++) {
		if (!(i % 32)) {
			sprintf(cmdstring, "@DE%03X", i);
			sendappcmd(fd, cmdstring);
			printf(".");
			fflush(stdout);
		}
		sprintf(cmdstring, "@X%04X%02X", i, data[i]);
		if (strcmp(sendappcmd(fd, cmdstring), cmdstring + 6)) {
			fprintf(stderr, "Error uploading byte %04x\n", i);
			return 0;
		}
	}
	printf("\nDeleting loader\n");
	if (!erasecard(fd)) {
		fprintf(stderr, "Could not erase EEPROM loader\n");
		return 0;
	}

	return 1;
}

int loadpic(int fd, unsigned char *progmem, unsigned char *datamem, unsigned short config)
{
	int i;
	unsigned char cmdbuf[255];
	MODE0;
	MODE2;

	printf("Programming PIC program memory ");

	for (i = 0; i < 0x800; i += 2) {
		if (!(i % 32)) {
			sprintf(cmdbuf, "@DP%03X", i >> 1);
			sendappcmd(fd, cmdbuf);
			printf(".");
			fflush(stdout);
		}
		sprintf(cmdbuf, "@L00%04X", *((unsigned short *) (progmem + i)));
		if (strcmp(sendappcmd(fd, cmdbuf), cmdbuf + 4)) {
			fprintf(stderr, "PIC Progmem Error at 0x%x\n", i);
			return 0;
		}
		sendappcmd(fd, "@I");
	}

	sendappcmd(fd, "@DDATA");
	printf("\nProgramming PIC data memory\n");

	for (i = 0; i < 0x80; i += 2) {
		sprintf(cmdbuf, "@L01%04X", *((unsigned short *) (datamem + i)));
		if (strcmp(sendappcmd(fd, cmdbuf) + 2, cmdbuf + 6)) {
			fprintf(stderr, "PIC Datamem Error at 0x%x\n", i);
			return 0;
		}
		sendappcmd(fd, "@I");
	}
	MODE0;
	MODE2;

	sendappcmd(fd, "@DCFG ");
	printf("Programming PIC config\n");

	sendappcmd(fd, "@P003FFF");
	for (i = 0; i < 4; i++) {
		sendappcmd(fd, "@L003FFF");
		sendappcmd(fd, "@I");
	}
	for (i = 0; i < 3; i++)
		sendappcmd(fd, "@I");
	sprintf(cmdbuf, "@L00%04X", config);
	sendappcmd(fd, cmdbuf);
	MODE0;
	return 1;
}

int main(int argc, char **argv)
{
	int fd, eefd, i;
	FILE *hexfd;
	FILE *eefdhex;
	struct termios oldtio, newtio;
	char buf[255];
	unsigned char eeprom[2048];
	unsigned char progmem[2048];
	unsigned char datamem[128];
	int address, bytes;
	unsigned short config = 0;
	unsigned char *bufptr;
	unsigned char *picfilename, *eepromfilename;
	int toggle = -1;

	for (i = 0; i < 1024; i++)
		*(((unsigned short *) progmem) + i) = 0x3fff;
	for (i = 0; i < 128; i++)
		*(((unsigned short *) datamem) + i) = 0x3fff;

	printf("MP2000 for Linux " VERSION "\n");
	printf("(C) 2000 Ulrich Hecht\n");
	printf("licensed under the terms of the GNU Public License (GPL) v2.\n");
	printf("EEPROM loader and Atmel software (C) ER Products\n\n");

	if (argc < 2) {
		printf("Usage: %s <PIC hex file> <EEPROM bin or hex file>\n", argv[0]);
		printf("using serial device %s\n", MODEMDEVICE);
		exit(1);
	}

	picfilename = argv[1];
	eepromfilename = argv[2];

#ifdef DEBUG_HEX
	goto hex;
#endif

	fd = open(MODEMDEVICE, O_RDWR | O_NOCTTY);
	if (fd < 0) {
		perror(MODEMDEVICE);
		exit(-1);
	}

	tcgetattr(fd, &oldtio);	/* save current port settings */

	bzero(&newtio, sizeof(newtio));
	newtio.c_cflag = BAUDRATE | CS8 | CLOCAL | CREAD;
	newtio.c_iflag = IGNPAR;
	newtio.c_oflag = 0;

	/* set input mode (non-canonical, no echo,...) */
	newtio.c_lflag = 0;

	newtio.c_cc[VTIME] = 0;	/* inter-character timer unused */
	newtio.c_cc[VMIN] = 1;	/* blocking read until 1 char received */

	tcflush(fd, TCIFLUSH);
	tcsetattr(fd, TCSANOW, &newtio);

	printf("Searching programmer... ");
	fflush(stdout);
	sendbioscmd(fd, "MPAT\r", ">");
	printf("found\n");

	if (!sendbioscmd(fd, "MPAT!\r", "MULTIPROG 2000\r\n>")) {
		fprintf(stderr, "This doesn't seem to be a Multiprog 2000\n");
		exit(1);
	}

	printf("Serial number %s\n", getbiosinfo(fd, "MPAT#\r"));
	printf("BIOS version %s\n\n", getbiosinfo(fd, "MPATV\r"));
	printf("Uploading software\n");

#include "apploader.h"

	sendbioscmd(fd, "MPAT\r", ">");
	printf("Starting software\n\n");
	sendbioscmd(fd, "MPAT\rMPAT%\r", ">");

	usleep(500);		/* the programmer needs a little time to start the software */

	while (!checkwafer(fd)) {
		if (toggle == -1) {
			toggle = 1;
			printf("Insert Goldwafer card");
		}
		if (toggle == 1) {
			sendappcmd(fd, "@DINS ");
			printf(".");
			fflush(stdout);
			toggle = !toggle;
		}
		else {
			sendappcmd(fd, "@DCARD");
			printf(".");
			fflush(stdout);
			toggle = !toggle;
		}
		sleep(1);
	}

	printf("\n\nErasing card\n\n");

	if (!erasecard(fd)) {
		fprintf(stderr, "Could not erase card\n");
		MODE0;
		exit(1);
	}

	if (strcmp(eepromfilename + strlen(eepromfilename) - 3, "hex") == 0
	    || strcmp(eepromfilename + strlen(eepromfilename) - 3, "HEX") == 0) {
		printf("Assuming EEPROM file %s to be an Intel HEX file\n", eepromfilename);
		eefdhex = fopen(eepromfilename, "r");
		if (eefdhex == NULL) {
			perror(eepromfilename);
			exit(1);
		}

		while (!feof(eefdhex)) {
			fgets(buf, 255, eefdhex);
			if (buf[0] != ':')
				continue;
			if (memcmp(buf + 7, "00", 2))
				continue;

			address = (a2h(buf[3]) << 12) + (a2h(buf[4]) << 8) + (a2h(buf[5]) << 4) + a2h(buf[6]);
			bytes = (a2h(buf[1]) << 4) + a2h(buf[2]);

#ifdef DEBUG_EEHEX
			printf("buf: -%s-\n", buf);
			printf("address: 0x%x\n", address);
			printf("bytes: %d\n", bytes);
#endif

			if (address > 0x7ff) {
				fprintf(stderr, "Invalid address 0x%x in EEPROM file %s\n", address, eepromfilename);
				exit(1);
			}

			for (i = 0; i < bytes; i++) {
				*(eeprom + i + address) = (a2h(buf[9 + i * 2]) << 4) + a2h(buf[10 + i * 2]);
			}

		}
		fclose(eefdhex);

	}
	else {
		printf("Assuming EEPROM file %s to be a raw binary file\n", eepromfilename);
		eefd = open(eepromfilename, O_RDWR);
		if (eefd < 0) {
			perror(eepromfilename);
			exit(1);
		}

		read(eefd, eeprom, 2048);
		close(eefd);
	}

	if (!loadeeprom(fd, eeprom)) {
		fprintf(stderr, "Uploading EEPROM data failed\n");
		MODE0;
		exit(1);
	}

	printf("\n");

#ifdef DEBUG_HEX
      hex:
#endif
	hexfd = fopen(picfilename, "r");
	if (hexfd == NULL) {
		perror(picfilename);
		exit(1);
	}

	while (!feof(hexfd)) {
		fgets(buf, 255, hexfd);
		if (buf[0] != ':')
			continue;
		if (memcmp(buf + 7, "00", 2))
			continue;

		address = (a2h(buf[3]) << 12) + (a2h(buf[4]) << 8) + (a2h(buf[5]) << 4) + a2h(buf[6]);
		bytes = (a2h(buf[1]) << 4) + a2h(buf[2]);

#ifdef DEBUG_HEX
		printf("buf: -%s-\n", buf);
		printf("address: 0x%x\n", address);
		printf("bytes: %d\n", bytes);
#endif

		bufptr = NULL;
		if (address <= 0x7ff)
			bufptr = progmem;
		if (address >= 0x4200 && address <= 0x4280)
			bufptr = datamem - 0x4200;
		if (address == 0x400e)	/* I suppose this is the PIC configuration */
			bufptr = ((unsigned char *) (&config)) - 0x400e;
		if (bufptr == NULL) {
			fprintf(stderr, "Warning: cannot deal with address 0x%x in hex file\n", address);
			continue;
		}
		for (i = 0; i < bytes; i++) {
			*(bufptr + i + address) = (a2h(buf[9 + i * 2]) << 4) + a2h(buf[10 + i * 2]);
		}

	}
	fclose(hexfd);
#ifdef DEBUG_HEX
	eefd = open("hexbin", O_RDWR | O_CREAT);
	write(eefd, progmem, 0x800);
	close(eefd);
	exit(1);
#endif

	if (!loadpic(fd, progmem, datamem, config)) {
		fprintf(stderr, "Failed to program PIC\n");
		MODE0;
		exit(1);
	}

	printf("\nGoldwafer card programmed successfully. Have fun!\n");

	tcsetattr(fd, TCSANOW, &oldtio);
	return 0;
}
